// Configuration and configuration file handling (implementation)

#include "lib.h"

//
// Global variables
//

config_t config;

const char *keys_names[PA_QUIT + 1] = {
  "null",
  "up",
  "down",
  "left",
  "right",
  "pgup",
  "pgdn",
  "home",
  "end",
  "open",
  "open_apropos",
  "open_whatis",
  "sp_open",
  "sp_apropos",
  "sp_whatis",
  "index",
  "back",
  "fwrd",
  "history",
  "toc",
  "search",
  "search_back",
  "search_next",
  "search_prev",
  "help",
  "quit"
};

const wchar_t *keys_help[PA_QUIT + 1] = {
  L"Do nothing",
  L"Scroll up one line",
  L"Scroll down one line",
  L"Scroll left one tab stop",
  L"Scroll right one tab stop",
  L"Scroll up one page",
  L"Scroll down one page",
  L"Go to page top",
  L"Go to page end",
  L"Open focused link",
  L"Perform apropos on the focused link",
  L"Perform whatis on the focused link",
  L"Open a manual page using a dialog",
  L"Perform apropos using a dialog",
  L"Perform whatis using a dialog",
  L"Go to index (home) page",
  L"Go back one step in history",
  L"Go forward one step in history",
  L"Show history menu",
  L"Show table of contents",
  L"Forward search",
  L"Backward search",
  L"Go to next search result",
  L"Go to previous search results",
  L"Show this help message",
  L"Exit the program"
};

//
// Helper macros and functions
//

// Helper of `conf_init()` `conf_handler()`. Initialize the values of all color
// pairs used for transitions.
#define set_transition_colours                                                 \
  config.colours.trans_mode_name = 100 * config.colours.stat_indic_mode.pair + \
                                   config.colours.stat_indic_name.pair;        \
  config.colours.trans_name_loc = 100 * config.colours.stat_indic_name.pair +  \
                                  config.colours.stat_indic_loc.pair;          \
  config.colours.trans_prompt_help =                                           \
      100 * config.colours.stat_input_prompt.pair +                            \
      config.colours.stat_input_help.pair;                                     \
  config.colours.trans_prompt_em =                                             \
      100 * config.colours.stat_input_prompt.pair +                            \
      config.colours.stat_input_em.pair;

// Helper of various `conf_...()` functions and methods. Return true if `src`
// is equal to 'true', 'yes', or '1'. Return false if it's equal to 'false',
// 'no', or '0'. Set `errno` to 0 if parsed successfully or to -1 if not.
bool wcs2bool(const wchar_t *src) {
  if (0 == wcscasecmp(src, L"TRUE") || 0 == wcscasecmp(src, L"YES") ||
      0 == wcscmp(src, L"1")) {
    errno = 0;
    return true;
  } else if (0 == wcscasecmp(src, L"FALSE") || 0 == wcscasecmp(src, L"NO") ||
             0 == wcscmp(src, L"0")) {
    errno = 0;
    return false;
  } else {
    errno = -1;
    return false;
  }
}

// Helper of various `conf_...()` functions and methods. Return `t_true` if
// `src` is equal to 'true', 'yes', or '1'. Return `t_false` if it's equal to
// 'false', 'no', or '0'. Return `t_auto` if it's equal to 'auto' or '-'. Set
// `errno` to 0 if parsed successfully or to -1 if not.
trit_t wcs2trit(const wchar_t *src) {
  if (0 == wcscasecmp(src, L"TRUE") || 0 == wcscasecmp(src, L"YES") ||
      0 == wcscmp(src, L"1")) {
    errno = 0;
    return t_true;
  } else if (0 == wcscasecmp(src, L"FALSE") || 0 == wcscasecmp(src, L"NO") ||
      0 == wcscmp(src, L"0")) {
    errno = 0;
    return t_false;
  } else if (0 == wcscasecmp(src, L"AUTO") || 0 == wcscmp(src, L"-")) {
    errno = 0;
    return t_auto;
  } else {
    errno = -1;
    return t_false;
  }
}

// Helper of various `conf_...()` functions and methods. Return the integer
// value held into `src` as an integer, and set `errno` to 0. If `src` cannot be
// parsed into an integer return `INT_MIN` and set `errno` to -1.
int wcs2int(const wchar_t *src) {
  int val, vals_read;
  vals_read = swscanf(src, L"%d", &val);
  if (1 != vals_read) {
    errno = -1;
    return INT_MIN;
  } else {
    errno = 0;
    return val;
  }
}

// Helper of various `conf_...()` functions and methods. Return the ncurses
// color that corresponds to `src`. Allowed values for `src` are 'black', 'red',
// 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', a positive integer, or
// '#rrggbb'. Return -1 in case of parse error, or -2 if the terminal cannot
// support such a color.
//
// NOTE: Setting `src` to a '#rrggbb' causes terminal colors to be re-defined.
// `wcs2colour(NULL)` will reset the re-defined colors index and return 0. It
// should be called at the beginning of `configure()`.
short wcs2colour(const wchar_t *src) {
  static short rgb_i = 255;    // index for re-defined terminal colors

  // If `src` is NULL, reset `rgb_i`
  if (NULL == src) {
    rgb_i = 255;
    return 0;
  }

  // Try to parse `src` as integer. If successful, return it.
  const int srci = wcs2int(src);
  if (errno == 0 && srci >= 0) {
    if (tcap.colours < 256) {
      // Terminal does not support 256 colors; return -2
      return -2;
    }
    return srci;
  }

  // If `src` matches a color name such as 'black', 'red', etc. return the
  // corresponding color.
  else if (0 == wcscasecmp(src, L"BLACK"))
    return COLOR_BLACK;
  else if (0 == wcscasecmp(src, L"RED"))
    return COLOR_RED;
  else if (0 == wcscasecmp(src, L"GREEN"))
    return COLOR_GREEN;
  else if (0 == wcscasecmp(src, L"YELLOW"))
    return COLOR_YELLOW;
  else if (0 == wcscasecmp(src, L"BLUE"))
    return COLOR_BLUE;
  else if (0 == wcscasecmp(src, L"MAGENTA"))
    return COLOR_MAGENTA;
  else if (0 == wcscasecmp(src, L"CYAN"))
    return COLOR_CYAN;
  else if (0 == wcscasecmp(src, L"WHITE"))
    return COLOR_WHITE;

  // Otherwise, try to parse `src` as '#rrggbb' and return the corresponding
  // color
  else {
    unsigned red = 0, green = 0, blue = 0; // New color values
    if (3 != swscanf(src, L"#%2x%2x%2x", &red, &green, &blue)) {
      // Unable to parse src; return -1
      return -1;
    } else {
      if (! tcap.rgb) {
        // Terminal does not support 256 colors that can be re-defined; return
        // -2
        return -2;
      }
      red = (red * 1000) / 256;
      green = (green * 1000) / 256;
      blue = (blue * 1000) / 256;
      init_color(rgb_i, red, green, blue);
      return rgb_i--;
    }
  }
}

// Helper of various `conf_...()` functions and methods. Return the ncurses key
// mapping that corresponds to `src`. Return -1 in case of parse error.
int wcs2ch(const wchar_t *src) {
  if (0 == wcscasecmp(src, L"KEY_UP"))
    return KEY_UP;
  else if (0 == wcscasecmp(src, L"KEY_DOWN"))
    return KEY_DOWN;
  else if (0 == wcscasecmp(src, L"KEY_LEFT"))
    return KEY_LEFT;
  else if (0 == wcscasecmp(src, L"KEY_RIGHT"))
    return KEY_RIGHT;
  else if (0 == wcscasecmp(src, L"KEY_PPAGE"))
    return KEY_PPAGE;
  else if (0 == wcscasecmp(src, L"KEY_NPAGE"))
    return KEY_NPAGE;
  else if (0 == wcscasecmp(src, L"KEY_HOME"))
    return KEY_HOME;
  else if (0 == wcscasecmp(src, L"KEY_END"))
    return KEY_END;
  else if (0 == wcscasecmp(src, L"ESC"))
    return '\e';
  else if (0 == wcscasecmp(src, L"KEY_BREAK"))
    return KEY_BREAK;
  else if (0 == wcscasecmp(src, L"ETX"))
    return 0x03;
  else if (0 == wcscasecmp(src, L"KEY_ENTER"))
    return KEY_ENTER;
  else if (0 == wcscasecmp(src, L"LF"))
    return '\n';
  else if (0 == wcscasecmp(src, L"KEY_BACKSPACE"))
    return KEY_BACKSPACE;
  else if (0 == wcscasecmp(src, L"BS"))
    return '\b';
  else if (0 == wcscasecmp(src, L"HT"))
    return '\t';
  else if (0 == wcscasecmp(src, L"SPACE"))
    return ' ';
  else if (0 == wcscasecmp(src, L"F1"))
    return KEY_F(1);
  else if (0 == wcscasecmp(src, L"F2"))
    return KEY_F(2);
  else if (0 == wcscasecmp(src, L"F3"))
    return KEY_F(3);
  else if (0 == wcscasecmp(src, L"F4"))
    return KEY_F(4);
  else if (0 == wcscasecmp(src, L"F5"))
    return KEY_F(5);
  else if (0 == wcscasecmp(src, L"F6"))
    return KEY_F(6);
  else if (0 == wcscasecmp(src, L"F7"))
    return KEY_F(7);
  else if (0 == wcscasecmp(src, L"F8"))
    return KEY_F(8);
  else if (0 == wcscasecmp(src, L"F9"))
    return KEY_F(9);
  else if (0 == wcscasecmp(src, L"F10"))
    return KEY_F(10);
  else if (0 == wcscasecmp(src, L"F11"))
    return KEY_F(11);
  else if (0 == wcscasecmp(src, L"F12"))
    return KEY_F(12);
  else if (src[0] >= 33 && src[0] <= 126 && 1 == wcslen(src))
    return src[0];
  else {
    // Unable to parse `src`; return -1
    return -1;
  }
}

// Helper of `conf_handler()`. Return an if statement that becomes true if
// `section` is equal to `s`.
#define conf_section(s) if (0 == wcscmp(section, s))

// Helper of `conf_handler()`. Set `trgt` to the boolean stored in `value`,
// using `wcs2bool()`, if `name` is equal to `n`.
#define conf_set_bool(n, trgt)                                                 \
  if (0 == wcscmp(name, n)) {                                                  \
    const bool val = wcs2bool(value);                                          \
    if (0 != errno) {                                                          \
      swprintf(errmsg, BS_SHORT,                                               \
               L"Error in %s line %d: Expected boolean, got '%ls'",            \
               path, line, value);                                             \
      winddown(ES_CONFIG_ERROR, errmsg);                                       \
    }                                                                          \
    trgt = val;                                                                \
    return;                                                                    \
  }

// Helper of `conf_handler()`. Set `trgt` to the `trit_t` stored in `value`,
// using `wcs2trit()`, if `name` is equal to `n`.
#define conf_set_trit(n, trgt)                                                 \
  if (0 == wcscmp(name, n)) {                                                  \
    const trit_t val = wcs2trit(value);                                        \
    if (0 != errno) {                                                          \
      swprintf(errmsg, BS_SHORT,                                               \
               L"Error in %s line %d: Expected ternary, got '%ls'",            \
               path, line, value);                                             \
      winddown(ES_CONFIG_ERROR, errmsg);                                       \
    }                                                                          \
    trgt = val;                                                                \
    return;                                                                    \
  }                                                                            \

// Helper of `conf_handler()`. Set `trgt` to the integer stored in `value`,
// using `wcs2int()`, if `name` is equal to `n`. `min` and `max` can be used to
// define a lower and an upper boundary for `value`, for error-checking
// purposes.
#define conf_set_int(n, trgt, min, max)                                        \
  if (0 == wcscmp(name, n)) {                                                  \
    const int val = wcs2int(value);                                            \
    if (0 != errno) {                                                          \
      swprintf(errmsg, BS_SHORT,                                               \
               L"Error in %s line %d: Expected integer, got '%ls'",            \
               path, line, value);                                             \
      winddown(ES_CONFIG_ERROR, errmsg);                                       \
    }                                                                          \
    if (val < min || val > max) {                                              \
      swprintf(errmsg, BS_SHORT,                                               \
               L"Error in %s line %d: Expected %d to %d, got %d",              \
               path, line, min, max, val);                                     \
      winddown(ES_CONFIG_ERROR, errmsg);                                       \
    }                                                                          \
    trgt = val;                                                                \
    return;                                                                    \
  }

// Helper of `conf_handler()`. Set `trgt` to a string copy of `value`, if `name`
// is equal to `n`.
#define conf_set_string(n, trgt)                                               \
  if (0 == wcscmp(name, n)) {                                                  \
    const unsigned value_len = wcslen(value);                                  \
    if (NULL != trgt)                                                          \
      free(trgt);                                                              \
    if (NULL == value)                                                         \
      trgt = NULL;                                                             \
    else {                                                                     \
      trgt = walloc(value_len * 4 + 1);                                        \
      xwcstombs(trgt, value, value_len * 4 + 1);                               \
    }                                                                          \
    return;                                                                    \
  }

// Helper of `conf_handler()`. Set `trgt` to a wide string copy of `value`, if
// `name` is equal to `n`.
#define conf_set_wstring(n, trgt)                                              \
  if (0 == wcscmp(name, n)) {                                                  \
    /* If the terminal doesn't support Unicode, reject all non-ASCII TUI    *  \
     * drawing characters                                                   */ \
    if (! tcap.unicode && 0 == wcscmp(section, L"chars") &&                    \
             (value[0] < 0 || value[0] > 127))                                 \
      return;                                                                  \
    if (NULL != trgt)                                                          \
      free(trgt);                                                              \
    if (NULL == value)                                                         \
      trgt = NULL;                                                             \
    else                                                                       \
      trgt = xwcsdup(value);                                                   \
    return;                                                                    \
  }

// Helper of `conf_handler()`. Set `trgt` to the color that corresponds to
// `value`, if `name` is equal to `n`. `value` is formatted as 2 color
// definitions (as accepted by `wcs2colour()`) and one boolean, separated by
// whitespace.
#define conf_set_colour(n, trgt)                                               \
  if (0 == wcscmp(name, n)) {                                                  \
    wchar_t fg[10], bg[10], bold[10];                                          \
    unsigned nfg, nbg;                                                         \
    bool nbold;                                                                \
    const int vals_read = swscanf(value, L"%9ls%9ls%9ls", fg, bg, bold);       \
    if (3 != vals_read) {                                                      \
      swprintf(errmsg, BS_SHORT,                                               \
               L"Error in %s line %d: Invalid color value '%ls'",              \
               path, line, value);                                             \
      winddown(ES_CONFIG_ERROR, errmsg);                                       \
    }                                                                          \
    nfg = wcs2colour(fg);                                                      \
    if (-1 == nfg) {                                                           \
      swprintf(                                                                \
          errmsg, BS_SHORT,                                                    \
          L"Error in %s line %d: Invalid foreground color '%ls'",              \
          path, line, fg);                                                     \
      winddown(ES_CONFIG_ERROR, errmsg);                                       \
    }                                                                          \
    nbg = wcs2colour(bg);                                                      \
    if (-1 == nbg) {                                                           \
      swprintf(                                                                \
          errmsg, BS_SHORT,                                                    \
          L"Error in %s line %d: Invalid background color '%ls'",              \
          path, line, bg);                                                     \
      winddown(ES_CONFIG_ERROR, errmsg);                                       \
    }                                                                          \
    nbold = wcs2bool(bold);                                                    \
    if (0 != errno) {                                                          \
      swprintf(errmsg, BS_SHORT,                                               \
               L"Error in %s line %d: Invalid boolean value '%ls'",            \
               path, line, bold);                                              \
      winddown(ES_CONFIG_ERROR, errmsg);                                       \
    }                                                                          \
    /* Apply colour changes only if the terminal supports it                */ \
    if (-2 != nfg && -2 != nbg) {                                              \
      trgt.fg = nfg;                                                           \
      trgt.bg = nbg;                                                           \
      trgt.bold = nbold;                                                       \
    }                                                                          \
    return;                                                                    \
  }

// Helper of `conf_handler()`. Set `trgt` to the set of key mappings that
// correspond to `value`, if `name` is equal to `n`. `value` is formatted as 8
// key mapping definitions (as accepted by `wcs2ch()`) separated by whitespace.
#define conf_set_key(n, trgt)                                                  \
  if (0 == wcscmp(name, n)) {                                                  \
    unsigned i;                                                                \
    wchar_t *keys[8];                                                          \
    for (i = 0; i < 8; i++)                                                    \
      keys[i] = walloca(BS_SHORT);                                             \
    const int keys_read = swscanf(value, L"%ls%ls%ls%ls%ls%ls%ls%ls", keys[0], \
                                 keys[1], keys[2], keys[3], keys[4], keys[5],  \
                                 keys[6], keys[7]);                            \
    if (0 == keys_read) {                                                      \
      swprintf(                                                                \
          errmsg, BS_SHORT,                                                    \
          L"Error in %s line %d: Invalid key mapping value '%ls'",             \
          path, line, value);                                                  \
      winddown(ES_CONFIG_ERROR, errmsg);                                       \
    }                                                                          \
    for (i = 0; i < keys_read; i++) {                                          \
      trgt[i] = wcs2ch(keys[i]);                                               \
      if (-1 == trgt[i]) {                                                     \
        swprintf(                                                              \
            errmsg, BS_SHORT,                                                  \
            L"Error in %s line %d: Key mapping '%ls' is invalid",              \
            path, line, keys[i]);                                              \
        winddown(ES_CONFIG_ERROR, errmsg);                                     \
      }                                                                        \
    }                                                                          \
    for (i = keys_read; i < 8; i++)                                            \
      trgt[i] = 0;                                                             \
    return;                                                                    \
  }

// Helper of `conf_handler()`. Error out with a "no such option" error message.
#define conf_no_such_option                                                    \
  swprintf(                                                                    \
      errmsg, BS_SHORT,                                                        \
      L"Error in %s line %d: No such option '%ls' in section '%ls'",           \
      path, line, name, section);                                              \
  winddown(ES_CONFIG_ERROR, errmsg);                                           \
  return;

// Helper of `conf_handler()`. Error out with a "no such section" error message.
#define conf_no_such_section                                                   \
  swprintf(errmsg, BS_SHORT,                                                   \
           L"Error in %s line %d: No such section '%ls'",                      \
           path, line, section);                                               \
  winddown(ES_CONFIG_ERROR, errmsg);                                           \
  return;

// Handler function for `eini()`
void conf_handler(const wchar_t *section, const wchar_t *name,
                  const wchar_t *value, const char *path, const unsigned line) {
  wchar_t errmsg[BS_SHORT]; // error message

conf_section(L"tcap") {
  conf_set_int(L"colours", config.tcap.colours, -1, 256);
  conf_set_trit(L"rgb", config.tcap.rgb);
  conf_set_trit(L"unicode", config.tcap.unicode);
  conf_set_trit(L"clipboard", config.tcap.clipboard);
  conf_set_int(L"escdelay", config.tcap.escdelay, 1, 10000);
  conf_no_such_option;
}

conf_section(L"chars") {
  conf_set_wstring(L"sbar_top", config.chars.sbar_top);
  conf_set_wstring(L"sbar_vline", config.chars.sbar_vline);
  conf_set_wstring(L"sbar_bottom", config.chars.sbar_bottom);
  conf_set_wstring(L"sbar_block", config.chars.sbar_block);
  conf_set_wstring(L"trans_mode_name", config.chars.trans_mode_name);
  conf_set_wstring(L"trans_name_loc", config.chars.trans_name_loc);
  conf_set_wstring(L"trans_prompt_help", config.chars.trans_prompt_help);
  conf_set_wstring(L"trans_prompt_em", config.chars.trans_prompt_em);
  conf_set_wstring(L"box_hline", config.chars.box_hline);
  conf_set_wstring(L"box_vline", config.chars.box_vline);
  conf_set_wstring(L"box_tl", config.chars.box_tl);
  conf_set_wstring(L"box_tr", config.chars.box_tr);
  conf_set_wstring(L"box_bl", config.chars.box_bl);
  conf_set_wstring(L"box_br", config.chars.box_br);
  conf_set_wstring(L"arrow_up", config.chars.arrow_up);
  conf_set_wstring(L"arrow_down", config.chars.arrow_down);
  conf_no_such_option;
}

conf_section(L"colours") {
  conf_set_colour(L"text", config.colours.text);
  conf_set_colour(L"search", config.colours.search);
  conf_set_colour(L"mark", config.colours.mark);
  conf_set_colour(L"link_man", config.colours.link_man);
  conf_set_colour(L"link_man_f", config.colours.link_man_f);
  conf_set_colour(L"link_http", config.colours.link_http);
  conf_set_colour(L"link_http_f", config.colours.link_http_f);
  conf_set_colour(L"link_email", config.colours.link_email);
  conf_set_colour(L"link_email_f", config.colours.link_email_f);
  conf_set_colour(L"link_ls", config.colours.link_ls);
  conf_set_colour(L"link_ls_f", config.colours.link_ls_f);
  conf_set_colour(L"sb_line", config.colours.sb_line);
  conf_set_colour(L"sb_block", config.colours.sb_block);
  conf_set_colour(L"stat_indic_mode", config.colours.stat_indic_mode);
  conf_set_colour(L"stat_indic_name", config.colours.stat_indic_name);
  conf_set_colour(L"stat_indic_loc", config.colours.stat_indic_loc);
  conf_set_colour(L"stat_input_prompt", config.colours.stat_input_prompt);
  conf_set_colour(L"stat_input_help", config.colours.stat_input_help);
  conf_set_colour(L"stat_input_em", config.colours.stat_input_em);
  conf_set_colour(L"imm_border", config.colours.imm_border);
  conf_set_colour(L"imm_title", config.colours.imm_title);
  conf_set_colour(L"sp_input", config.colours.sp_input);
  conf_set_colour(L"sp_text", config.colours.sp_text);
  conf_set_colour(L"sp_text_f", config.colours.sp_text_f);
  conf_set_colour(L"help_text", config.colours.help_text);
  conf_set_colour(L"help_text_f", config.colours.help_text_f);
  conf_set_colour(L"history_text", config.colours.history_text);
  conf_set_colour(L"history_text_f", config.colours.history_text_f);
  conf_set_colour(L"toc_text", config.colours.toc_text);
  conf_set_colour(L"toc_text_f", config.colours.toc_text_f);
  conf_no_such_option;
}

conf_section(L"keys") {
  conf_set_key(L"up", config.keys[PA_UP]);
  conf_set_key(L"down", config.keys[PA_DOWN]);
  conf_set_key(L"left", config.keys[PA_LEFT]);
  conf_set_key(L"right", config.keys[PA_RIGHT]);
  conf_set_key(L"pgup", config.keys[PA_PGUP]);
  conf_set_key(L"pgdn", config.keys[PA_PGDN]);
  conf_set_key(L"home", config.keys[PA_HOME]);
  conf_set_key(L"end", config.keys[PA_END]);
  conf_set_key(L"open", config.keys[PA_OPEN]);
  conf_set_key(L"open_apropos", config.keys[PA_OPEN_APROPOS]);
  conf_set_key(L"open_whatis", config.keys[PA_OPEN_WHATIS]);
  conf_set_key(L"sp_open", config.keys[PA_SP_OPEN]);
  conf_set_key(L"sp_apropos", config.keys[PA_SP_APROPOS]);
  conf_set_key(L"sp_whatis", config.keys[PA_SP_WHATIS]);
  conf_set_key(L"index", config.keys[PA_INDEX]);
  conf_set_key(L"back", config.keys[PA_BACK]);
  conf_set_key(L"fwrd", config.keys[PA_FWRD]);
  conf_set_key(L"history", config.keys[PA_HISTORY]);
  conf_set_key(L"toc", config.keys[PA_TOC]);
  conf_set_key(L"search", config.keys[PA_SEARCH]);
  conf_set_key(L"search_back", config.keys[PA_SEARCH_BACK]);
  conf_set_key(L"search_next", config.keys[PA_SEARCH_NEXT]);
  conf_set_key(L"search_prev", config.keys[PA_SEARCH_PREV]);
  conf_set_key(L"help", config.keys[PA_HELP]);
  conf_set_key(L"quit", config.keys[PA_QUIT]);
  conf_no_such_option;
}

conf_section(L"mouse") {
  conf_set_bool(L"enable", config.mouse.enable);
  conf_set_bool(L"left_handed", config.mouse.left_handed);
  conf_set_bool(L"left_click_open", config.mouse.left_click_open);
  conf_no_such_option;
}

conf_section(L"layout") {
  conf_set_bool(L"sections_on_top", config.layout.sections_on_top);
  conf_set_int(L"lmargin", config.layout.lmargin, 0, 200);
  conf_set_int(L"rmargin", config.layout.rmargin, 0, 200);
  conf_set_int(L"tabstop", config.layout.tabstop, 0, 100);
  conf_set_bool(L"sbar", config.layout.sbar);
  conf_set_bool(L"beep", config.layout.beep);
  conf_no_such_option;
}

conf_section(L"misc") {
  conf_set_string(L"man_path", config.misc.man_path);
  conf_set_string(L"groff_path", config.misc.groff_path);
  conf_set_string(L"whatis_path", config.misc.whatis_path);
  conf_set_string(L"apropos_path", config.misc.apropos_path);
  conf_set_string(L"browser_path", config.misc.browser_path);
  conf_set_string(L"mailer_path", config.misc.mailer_path);
  conf_set_bool(L"reset_after_http", config.misc.reset_after_http);
  conf_set_bool(L"reset_after_email", config.misc.reset_after_email);
  conf_set_int(L"history_size", config.misc.history_size, 0, 262144);
  conf_set_bool(L"hyphenate", config.misc.hyphenate);
  conf_set_bool(L"justify", config.misc.justify);
  conf_no_such_option;
}


  conf_no_such_section;
}

// Error function for `eini()`
void error_handler(const wchar_t *error, const char *path,
                   const unsigned line) {
  wchar_t errmsg[BS_SHORT]; // error message

  swprintf(errmsg, BS_SHORT,
           L"Error in %s line %d: %ls",
           path, line, error);
  winddown(ES_CONFIG_ERROR, errmsg);
}

// Helper of `configure()`. Initialize configuration in sections `[chars]` and
// `[colours]` to sane defaults.
void conf_init_chars_colours() {
// Default value for: Scrollbar top end
if (NULL != config.chars.sbar_top)
  free(config.chars.sbar_top);
config.chars.sbar_top = xwcsdup(L"┬");

// Default value for: Scrollbar track line
if (NULL != config.chars.sbar_vline)
  free(config.chars.sbar_vline);
config.chars.sbar_vline = xwcsdup(L"│");

// Default value for: Scrollbar bottom end
if (NULL != config.chars.sbar_bottom)
  free(config.chars.sbar_bottom);
config.chars.sbar_bottom = xwcsdup(L"┴");

// Default value for: Scrollbar knob
if (NULL != config.chars.sbar_block)
  free(config.chars.sbar_block);
config.chars.sbar_block = xwcsdup(L"█");

// Default value for: Transition between the mode to name sections of the status bar
if (NULL != config.chars.trans_mode_name)
  free(config.chars.trans_mode_name);
config.chars.trans_mode_name = xwcsdup(L"│");

// Default value for: Transition between the name and location sections of the status bar
if (NULL != config.chars.trans_name_loc)
  free(config.chars.trans_name_loc);
config.chars.trans_name_loc = xwcsdup(L"│");

// Default value for: Transition between the prompt and help sections of the status bar
if (NULL != config.chars.trans_prompt_help)
  free(config.chars.trans_prompt_help);
config.chars.trans_prompt_help = xwcsdup(L" ");

// Default value for: Transition between the prompt and error message sections of the status bar
if (NULL != config.chars.trans_prompt_em)
  free(config.chars.trans_prompt_em);
config.chars.trans_prompt_em = xwcsdup(L" ");

// Default value for: Dialog box horizontal line
if (NULL != config.chars.box_hline)
  free(config.chars.box_hline);
config.chars.box_hline = xwcsdup(L"─");

// Default value for: Dialog box vertical line
if (NULL != config.chars.box_vline)
  free(config.chars.box_vline);
config.chars.box_vline = xwcsdup(L"│");

// Default value for: Dialog box top left corner
if (NULL != config.chars.box_tl)
  free(config.chars.box_tl);
config.chars.box_tl = xwcsdup(L"┌");

// Default value for: Dialog box top right corner
if (NULL != config.chars.box_tr)
  free(config.chars.box_tr);
config.chars.box_tr = xwcsdup(L"┐");

// Default value for: Dialog box bottom left corner
if (NULL != config.chars.box_bl)
  free(config.chars.box_bl);
config.chars.box_bl = xwcsdup(L"└");

// Default value for: Dialog box bottom right corner
if (NULL != config.chars.box_br)
  free(config.chars.box_br);
config.chars.box_br = xwcsdup(L"┘");

// Default value for: Up arrow
if (NULL != config.chars.arrow_up)
  free(config.chars.arrow_up);
config.chars.arrow_up = xwcsdup(L"↑");

// Default value for: Down arrow
if (NULL != config.chars.arrow_down)
  free(config.chars.arrow_down);
config.chars.arrow_down = xwcsdup(L"↓");

// Default value for: Fallback for B&W terminals
config.colours.fallback.fg = wcs2colour(L"white");
config.colours.fallback.bg = wcs2colour(L"black");
config.colours.fallback.bold = wcs2bool(L"false");
config.colours.fallback.pair = 10;

// Default value for: Page text
config.colours.text.fg = wcs2colour(L"white");
config.colours.text.bg = wcs2colour(L"black");
config.colours.text.bold = wcs2bool(L"false");
config.colours.text.pair = 11;

// Default value for: Matched search terms in page text
config.colours.search.fg = wcs2colour(L"black");
config.colours.search.bg = wcs2colour(L"white");
config.colours.search.bold = wcs2bool(L"false");
config.colours.search.pair = 12;

// Default value for: Marked text
config.colours.mark.fg = wcs2colour(L"white");
config.colours.mark.bg = wcs2colour(L"cyan");
config.colours.mark.bold = wcs2bool(L"false");
config.colours.mark.pair = 13;

// Default value for: Links to manual pages
config.colours.link_man.fg = wcs2colour(L"green");
config.colours.link_man.bg = wcs2colour(L"black");
config.colours.link_man.bold = wcs2bool(L"false");
config.colours.link_man.pair = 14;

// Default value for: Links to manual pages (focused)
config.colours.link_man_f.fg = wcs2colour(L"black");
config.colours.link_man_f.bg = wcs2colour(L"green");
config.colours.link_man_f.bold = wcs2bool(L"false");
config.colours.link_man_f.pair = 15;

// Default value for: HTTP links
config.colours.link_http.fg = wcs2colour(L"magenta");
config.colours.link_http.bg = wcs2colour(L"black");
config.colours.link_http.bold = wcs2bool(L"false");
config.colours.link_http.pair = 16;

// Default value for: HTTP links (focused)
config.colours.link_http_f.fg = wcs2colour(L"black");
config.colours.link_http_f.bg = wcs2colour(L"magenta");
config.colours.link_http_f.bold = wcs2bool(L"false");
config.colours.link_http_f.pair = 17;

// Default value for: E-mail links
config.colours.link_email.fg = wcs2colour(L"magenta");
config.colours.link_email.bg = wcs2colour(L"black");
config.colours.link_email.bold = wcs2bool(L"false");
config.colours.link_email.pair = 18;

// Default value for: E-mail links (focused)
config.colours.link_email_f.fg = wcs2colour(L"black");
config.colours.link_email_f.bg = wcs2colour(L"magenta");
config.colours.link_email_f.bold = wcs2bool(L"false");
config.colours.link_email_f.pair = 19;

// Default value for: In-page local search links
config.colours.link_ls.fg = wcs2colour(L"cyan");
config.colours.link_ls.bg = wcs2colour(L"black");
config.colours.link_ls.bold = wcs2bool(L"false");
config.colours.link_ls.pair = 20;

// Default value for: In-page local search links (focused)
config.colours.link_ls_f.fg = wcs2colour(L"black");
config.colours.link_ls_f.bg = wcs2colour(L"cyan");
config.colours.link_ls_f.bold = wcs2bool(L"false");
config.colours.link_ls_f.pair = 21;

// Default value for: Scrollbar track line
config.colours.sb_line.fg = wcs2colour(L"yellow");
config.colours.sb_line.bg = wcs2colour(L"black");
config.colours.sb_line.bold = wcs2bool(L"true");
config.colours.sb_line.pair = 22;

// Default value for: Scrollbar knob
config.colours.sb_block.fg = wcs2colour(L"yellow");
config.colours.sb_block.bg = wcs2colour(L"black");
config.colours.sb_block.bold = wcs2bool(L"true");
config.colours.sb_block.pair = 23;

// Default value for: Status bar mode section
config.colours.stat_indic_mode.fg = wcs2colour(L"yellow");
config.colours.stat_indic_mode.bg = wcs2colour(L"red");
config.colours.stat_indic_mode.bold = wcs2bool(L"true");
config.colours.stat_indic_mode.pair = 24;

// Default value for: Status bar name section
config.colours.stat_indic_name.fg = wcs2colour(L"white");
config.colours.stat_indic_name.bg = wcs2colour(L"blue");
config.colours.stat_indic_name.bold = wcs2bool(L"true");
config.colours.stat_indic_name.pair = 25;

// Default value for: Status bar location section
config.colours.stat_indic_loc.fg = wcs2colour(L"black");
config.colours.stat_indic_loc.bg = wcs2colour(L"white");
config.colours.stat_indic_loc.bold = wcs2bool(L"false");
config.colours.stat_indic_loc.pair = 26;

// Default value for: Status bar input prompt
config.colours.stat_input_prompt.fg = wcs2colour(L"white");
config.colours.stat_input_prompt.bg = wcs2colour(L"black");
config.colours.stat_input_prompt.bold = wcs2bool(L"false");
config.colours.stat_input_prompt.pair = 27;

// Default value for: Status bar help section
config.colours.stat_input_help.fg = wcs2colour(L"yellow");
config.colours.stat_input_help.bg = wcs2colour(L"black");
config.colours.stat_input_help.bold = wcs2bool(L"true");
config.colours.stat_input_help.pair = 28;

// Default value for: Status bar error message section
config.colours.stat_input_em.fg = wcs2colour(L"red");
config.colours.stat_input_em.bg = wcs2colour(L"black");
config.colours.stat_input_em.bold = wcs2bool(L"true");
config.colours.stat_input_em.pair = 29;

// Default value for: Pop-up dialogs border
config.colours.imm_border.fg = wcs2colour(L"yellow");
config.colours.imm_border.bg = wcs2colour(L"black");
config.colours.imm_border.bold = wcs2bool(L"true");
config.colours.imm_border.pair = 30;

// Default value for: Pop-up dialogs title
config.colours.imm_title.fg = wcs2colour(L"yellow");
config.colours.imm_title.bg = wcs2colour(L"red");
config.colours.imm_title.bold = wcs2bool(L"true");
config.colours.imm_title.pair = 31;

// Default value for: Pop-up input dialog prompt
config.colours.sp_input.fg = wcs2colour(L"white");
config.colours.sp_input.bg = wcs2colour(L"black");
config.colours.sp_input.bold = wcs2bool(L"false");
config.colours.sp_input.pair = 32;

// Default value for: Pop-up input dialog progressive help text
config.colours.sp_text.fg = wcs2colour(L"cyan");
config.colours.sp_text.bg = wcs2colour(L"black");
config.colours.sp_text.bold = wcs2bool(L"false");
config.colours.sp_text.pair = 33;

// Default value for: Pop-up input dialog progressive help text (focused)
config.colours.sp_text_f.fg = wcs2colour(L"white");
config.colours.sp_text_f.bg = wcs2colour(L"black");
config.colours.sp_text_f.bold = wcs2bool(L"false");
config.colours.sp_text_f.pair = 34;

// Default value for: Help dialog entries text
config.colours.help_text.fg = wcs2colour(L"cyan");
config.colours.help_text.bg = wcs2colour(L"black");
config.colours.help_text.bold = wcs2bool(L"false");
config.colours.help_text.pair = 35;

// Default value for: Help dialog entries text (focused)
config.colours.help_text_f.fg = wcs2colour(L"black");
config.colours.help_text_f.bg = wcs2colour(L"cyan");
config.colours.help_text_f.bold = wcs2bool(L"false");
config.colours.help_text_f.pair = 36;

// Default value for: History dialog entries text
config.colours.history_text.fg = wcs2colour(L"cyan");
config.colours.history_text.bg = wcs2colour(L"black");
config.colours.history_text.bold = wcs2bool(L"false");
config.colours.history_text.pair = 37;

// Default value for: History dialog entries text (focused)
config.colours.history_text_f.fg = wcs2colour(L"black");
config.colours.history_text_f.bg = wcs2colour(L"cyan");
config.colours.history_text_f.bold = wcs2bool(L"false");
config.colours.history_text_f.pair = 38;

// Default value for: TOC dialog entries text
config.colours.toc_text.fg = wcs2colour(L"cyan");
config.colours.toc_text.bg = wcs2colour(L"black");
config.colours.toc_text.bold = wcs2bool(L"false");
config.colours.toc_text.pair = 39;

// Default value for: TOC dialog entries text (focused)
config.colours.toc_text_f.fg = wcs2colour(L"black");
config.colours.toc_text_f.bg = wcs2colour(L"cyan");
config.colours.toc_text_f.bold = wcs2bool(L"false");
config.colours.toc_text_f.pair = 40;

// Default value for: Color pair for mode to name transition character
config.colours.trans_mode_name = wcs2int(L"0");

// Default value for: Color pair for name to location transition character
config.colours.trans_name_loc = wcs2int(L"0");

// Default value for: Color pair for prompt to help transition character
config.colours.trans_prompt_help = wcs2int(L"0");

// Default value for: Color pair for prompt to error_message transition character
config.colours.trans_prompt_em = wcs2int(L"0");


  set_transition_colours;
}

//
// Functions
//

void conf_init() {
// Default value for: Number of colors supported by the terminal (set to -1 to auto-configure)
config.tcap.colours = wcs2int(L"-1");

// Default value for: Terminal colors can be re-defined
config.tcap.rgb = wcs2trit(L"auto");

// Default value for: Terminal supports Unicode
config.tcap.unicode = wcs2trit(L"auto");

// Default value for: Terminal supports clipboard operations (OSC 52)
config.tcap.clipboard = wcs2trit(L"auto");

// Default value for: Terminal escape delay
config.tcap.escdelay = wcs2int(L"60");

// Default value for: Scrollbar top end
config.chars.sbar_top = xwcsdup(L"┬");

// Default value for: Scrollbar track line
config.chars.sbar_vline = xwcsdup(L"│");

// Default value for: Scrollbar bottom end
config.chars.sbar_bottom = xwcsdup(L"┴");

// Default value for: Scrollbar knob
config.chars.sbar_block = xwcsdup(L"█");

// Default value for: Transition between the mode to name sections of the status bar
config.chars.trans_mode_name = xwcsdup(L"│");

// Default value for: Transition between the name and location sections of the status bar
config.chars.trans_name_loc = xwcsdup(L"│");

// Default value for: Transition between the prompt and help sections of the status bar
config.chars.trans_prompt_help = xwcsdup(L" ");

// Default value for: Transition between the prompt and error message sections of the status bar
config.chars.trans_prompt_em = xwcsdup(L" ");

// Default value for: Dialog box horizontal line
config.chars.box_hline = xwcsdup(L"─");

// Default value for: Dialog box vertical line
config.chars.box_vline = xwcsdup(L"│");

// Default value for: Dialog box top left corner
config.chars.box_tl = xwcsdup(L"┌");

// Default value for: Dialog box top right corner
config.chars.box_tr = xwcsdup(L"┐");

// Default value for: Dialog box bottom left corner
config.chars.box_bl = xwcsdup(L"└");

// Default value for: Dialog box bottom right corner
config.chars.box_br = xwcsdup(L"┘");

// Default value for: Up arrow
config.chars.arrow_up = xwcsdup(L"↑");

// Default value for: Down arrow
config.chars.arrow_down = xwcsdup(L"↓");

// Default value for: Fallback for B&W terminals
config.colours.fallback.fg = wcs2colour(L"white");
config.colours.fallback.bg = wcs2colour(L"black");
config.colours.fallback.bold = wcs2bool(L"false");
config.colours.fallback.pair = 10;

// Default value for: Page text
config.colours.text.fg = wcs2colour(L"white");
config.colours.text.bg = wcs2colour(L"black");
config.colours.text.bold = wcs2bool(L"false");
config.colours.text.pair = 11;

// Default value for: Matched search terms in page text
config.colours.search.fg = wcs2colour(L"black");
config.colours.search.bg = wcs2colour(L"white");
config.colours.search.bold = wcs2bool(L"false");
config.colours.search.pair = 12;

// Default value for: Marked text
config.colours.mark.fg = wcs2colour(L"white");
config.colours.mark.bg = wcs2colour(L"cyan");
config.colours.mark.bold = wcs2bool(L"false");
config.colours.mark.pair = 13;

// Default value for: Links to manual pages
config.colours.link_man.fg = wcs2colour(L"green");
config.colours.link_man.bg = wcs2colour(L"black");
config.colours.link_man.bold = wcs2bool(L"false");
config.colours.link_man.pair = 14;

// Default value for: Links to manual pages (focused)
config.colours.link_man_f.fg = wcs2colour(L"black");
config.colours.link_man_f.bg = wcs2colour(L"green");
config.colours.link_man_f.bold = wcs2bool(L"false");
config.colours.link_man_f.pair = 15;

// Default value for: HTTP links
config.colours.link_http.fg = wcs2colour(L"magenta");
config.colours.link_http.bg = wcs2colour(L"black");
config.colours.link_http.bold = wcs2bool(L"false");
config.colours.link_http.pair = 16;

// Default value for: HTTP links (focused)
config.colours.link_http_f.fg = wcs2colour(L"black");
config.colours.link_http_f.bg = wcs2colour(L"magenta");
config.colours.link_http_f.bold = wcs2bool(L"false");
config.colours.link_http_f.pair = 17;

// Default value for: E-mail links
config.colours.link_email.fg = wcs2colour(L"magenta");
config.colours.link_email.bg = wcs2colour(L"black");
config.colours.link_email.bold = wcs2bool(L"false");
config.colours.link_email.pair = 18;

// Default value for: E-mail links (focused)
config.colours.link_email_f.fg = wcs2colour(L"black");
config.colours.link_email_f.bg = wcs2colour(L"magenta");
config.colours.link_email_f.bold = wcs2bool(L"false");
config.colours.link_email_f.pair = 19;

// Default value for: In-page local search links
config.colours.link_ls.fg = wcs2colour(L"cyan");
config.colours.link_ls.bg = wcs2colour(L"black");
config.colours.link_ls.bold = wcs2bool(L"false");
config.colours.link_ls.pair = 20;

// Default value for: In-page local search links (focused)
config.colours.link_ls_f.fg = wcs2colour(L"black");
config.colours.link_ls_f.bg = wcs2colour(L"cyan");
config.colours.link_ls_f.bold = wcs2bool(L"false");
config.colours.link_ls_f.pair = 21;

// Default value for: Scrollbar track line
config.colours.sb_line.fg = wcs2colour(L"yellow");
config.colours.sb_line.bg = wcs2colour(L"black");
config.colours.sb_line.bold = wcs2bool(L"true");
config.colours.sb_line.pair = 22;

// Default value for: Scrollbar knob
config.colours.sb_block.fg = wcs2colour(L"yellow");
config.colours.sb_block.bg = wcs2colour(L"black");
config.colours.sb_block.bold = wcs2bool(L"true");
config.colours.sb_block.pair = 23;

// Default value for: Status bar mode section
config.colours.stat_indic_mode.fg = wcs2colour(L"yellow");
config.colours.stat_indic_mode.bg = wcs2colour(L"red");
config.colours.stat_indic_mode.bold = wcs2bool(L"true");
config.colours.stat_indic_mode.pair = 24;

// Default value for: Status bar name section
config.colours.stat_indic_name.fg = wcs2colour(L"white");
config.colours.stat_indic_name.bg = wcs2colour(L"blue");
config.colours.stat_indic_name.bold = wcs2bool(L"true");
config.colours.stat_indic_name.pair = 25;

// Default value for: Status bar location section
config.colours.stat_indic_loc.fg = wcs2colour(L"black");
config.colours.stat_indic_loc.bg = wcs2colour(L"white");
config.colours.stat_indic_loc.bold = wcs2bool(L"false");
config.colours.stat_indic_loc.pair = 26;

// Default value for: Status bar input prompt
config.colours.stat_input_prompt.fg = wcs2colour(L"white");
config.colours.stat_input_prompt.bg = wcs2colour(L"black");
config.colours.stat_input_prompt.bold = wcs2bool(L"false");
config.colours.stat_input_prompt.pair = 27;

// Default value for: Status bar help section
config.colours.stat_input_help.fg = wcs2colour(L"yellow");
config.colours.stat_input_help.bg = wcs2colour(L"black");
config.colours.stat_input_help.bold = wcs2bool(L"true");
config.colours.stat_input_help.pair = 28;

// Default value for: Status bar error message section
config.colours.stat_input_em.fg = wcs2colour(L"red");
config.colours.stat_input_em.bg = wcs2colour(L"black");
config.colours.stat_input_em.bold = wcs2bool(L"true");
config.colours.stat_input_em.pair = 29;

// Default value for: Pop-up dialogs border
config.colours.imm_border.fg = wcs2colour(L"yellow");
config.colours.imm_border.bg = wcs2colour(L"black");
config.colours.imm_border.bold = wcs2bool(L"true");
config.colours.imm_border.pair = 30;

// Default value for: Pop-up dialogs title
config.colours.imm_title.fg = wcs2colour(L"yellow");
config.colours.imm_title.bg = wcs2colour(L"red");
config.colours.imm_title.bold = wcs2bool(L"true");
config.colours.imm_title.pair = 31;

// Default value for: Pop-up input dialog prompt
config.colours.sp_input.fg = wcs2colour(L"white");
config.colours.sp_input.bg = wcs2colour(L"black");
config.colours.sp_input.bold = wcs2bool(L"false");
config.colours.sp_input.pair = 32;

// Default value for: Pop-up input dialog progressive help text
config.colours.sp_text.fg = wcs2colour(L"cyan");
config.colours.sp_text.bg = wcs2colour(L"black");
config.colours.sp_text.bold = wcs2bool(L"false");
config.colours.sp_text.pair = 33;

// Default value for: Pop-up input dialog progressive help text (focused)
config.colours.sp_text_f.fg = wcs2colour(L"white");
config.colours.sp_text_f.bg = wcs2colour(L"black");
config.colours.sp_text_f.bold = wcs2bool(L"false");
config.colours.sp_text_f.pair = 34;

// Default value for: Help dialog entries text
config.colours.help_text.fg = wcs2colour(L"cyan");
config.colours.help_text.bg = wcs2colour(L"black");
config.colours.help_text.bold = wcs2bool(L"false");
config.colours.help_text.pair = 35;

// Default value for: Help dialog entries text (focused)
config.colours.help_text_f.fg = wcs2colour(L"black");
config.colours.help_text_f.bg = wcs2colour(L"cyan");
config.colours.help_text_f.bold = wcs2bool(L"false");
config.colours.help_text_f.pair = 36;

// Default value for: History dialog entries text
config.colours.history_text.fg = wcs2colour(L"cyan");
config.colours.history_text.bg = wcs2colour(L"black");
config.colours.history_text.bold = wcs2bool(L"false");
config.colours.history_text.pair = 37;

// Default value for: History dialog entries text (focused)
config.colours.history_text_f.fg = wcs2colour(L"black");
config.colours.history_text_f.bg = wcs2colour(L"cyan");
config.colours.history_text_f.bold = wcs2bool(L"false");
config.colours.history_text_f.pair = 38;

// Default value for: TOC dialog entries text
config.colours.toc_text.fg = wcs2colour(L"cyan");
config.colours.toc_text.bg = wcs2colour(L"black");
config.colours.toc_text.bold = wcs2bool(L"false");
config.colours.toc_text.pair = 39;

// Default value for: TOC dialog entries text (focused)
config.colours.toc_text_f.fg = wcs2colour(L"black");
config.colours.toc_text_f.bg = wcs2colour(L"cyan");
config.colours.toc_text_f.bold = wcs2bool(L"false");
config.colours.toc_text_f.pair = 40;

// Default value for: Color pair for mode to name transition character
config.colours.trans_mode_name = wcs2int(L"0");

// Default value for: Color pair for name to location transition character
config.colours.trans_name_loc = wcs2int(L"0");

// Default value for: Color pair for prompt to help transition character
config.colours.trans_prompt_help = wcs2int(L"0");

// Default value for: Color pair for prompt to error_message transition character
config.colours.trans_prompt_em = wcs2int(L"0");

// Default value for: Scroll up one line
config.keys[PA_UP][0] = wcs2ch(L"KEY_UP");
config.keys[PA_UP][1] = wcs2ch(L"y");
config.keys[PA_UP][2] = wcs2ch(L"k");
config.keys[PA_UP][3] = 0;
config.keys[PA_UP][4] = 0;
config.keys[PA_UP][5] = 0;
config.keys[PA_UP][6] = 0;
config.keys[PA_UP][7] = 0;

// Default value for: Scroll down one line
config.keys[PA_DOWN][0] = wcs2ch(L"KEY_DOWN");
config.keys[PA_DOWN][1] = wcs2ch(L"e");
config.keys[PA_DOWN][2] = wcs2ch(L"j");
config.keys[PA_DOWN][3] = 0;
config.keys[PA_DOWN][4] = 0;
config.keys[PA_DOWN][5] = 0;
config.keys[PA_DOWN][6] = 0;
config.keys[PA_DOWN][7] = 0;

// Default value for: Scroll left one tab stop
config.keys[PA_LEFT][0] = wcs2ch(L"KEY_LEFT");
config.keys[PA_LEFT][1] = wcs2ch(L"<");
config.keys[PA_LEFT][2] = 0;
config.keys[PA_LEFT][3] = 0;
config.keys[PA_LEFT][4] = 0;
config.keys[PA_LEFT][5] = 0;
config.keys[PA_LEFT][6] = 0;
config.keys[PA_LEFT][7] = 0;

// Default value for: Scroll right one tab stop
config.keys[PA_RIGHT][0] = wcs2ch(L"KEY_RIGHT");
config.keys[PA_RIGHT][1] = wcs2ch(L">");
config.keys[PA_RIGHT][2] = 0;
config.keys[PA_RIGHT][3] = 0;
config.keys[PA_RIGHT][4] = 0;
config.keys[PA_RIGHT][5] = 0;
config.keys[PA_RIGHT][6] = 0;
config.keys[PA_RIGHT][7] = 0;

// Default value for: Scroll up one page
config.keys[PA_PGUP][0] = wcs2ch(L"KEY_PPAGE");
config.keys[PA_PGUP][1] = wcs2ch(L"b");
config.keys[PA_PGUP][2] = 0;
config.keys[PA_PGUP][3] = 0;
config.keys[PA_PGUP][4] = 0;
config.keys[PA_PGUP][5] = 0;
config.keys[PA_PGUP][6] = 0;
config.keys[PA_PGUP][7] = 0;

// Default value for: Scroll down one page
config.keys[PA_PGDN][0] = wcs2ch(L"KEY_NPAGE");
config.keys[PA_PGDN][1] = wcs2ch(L"f");
config.keys[PA_PGDN][2] = 0;
config.keys[PA_PGDN][3] = 0;
config.keys[PA_PGDN][4] = 0;
config.keys[PA_PGDN][5] = 0;
config.keys[PA_PGDN][6] = 0;
config.keys[PA_PGDN][7] = 0;

// Default value for: Go to page top
config.keys[PA_HOME][0] = wcs2ch(L"KEY_HOME");
config.keys[PA_HOME][1] = wcs2ch(L"g");
config.keys[PA_HOME][2] = 0;
config.keys[PA_HOME][3] = 0;
config.keys[PA_HOME][4] = 0;
config.keys[PA_HOME][5] = 0;
config.keys[PA_HOME][6] = 0;
config.keys[PA_HOME][7] = 0;

// Default value for: Go to page end
config.keys[PA_END][0] = wcs2ch(L"KEY_END");
config.keys[PA_END][1] = wcs2ch(L"G");
config.keys[PA_END][2] = 0;
config.keys[PA_END][3] = 0;
config.keys[PA_END][4] = 0;
config.keys[PA_END][5] = 0;
config.keys[PA_END][6] = 0;
config.keys[PA_END][7] = 0;

// Default value for: Open focused link
config.keys[PA_OPEN][0] = wcs2ch(L"KEY_ENTER");
config.keys[PA_OPEN][1] = wcs2ch(L"LF");
config.keys[PA_OPEN][2] = wcs2ch(L"o");
config.keys[PA_OPEN][3] = 0;
config.keys[PA_OPEN][4] = 0;
config.keys[PA_OPEN][5] = 0;
config.keys[PA_OPEN][6] = 0;
config.keys[PA_OPEN][7] = 0;

// Default value for: Perform apropos on the focused link
config.keys[PA_OPEN_APROPOS][0] = wcs2ch(L"a");
config.keys[PA_OPEN_APROPOS][1] = 0;
config.keys[PA_OPEN_APROPOS][2] = 0;
config.keys[PA_OPEN_APROPOS][3] = 0;
config.keys[PA_OPEN_APROPOS][4] = 0;
config.keys[PA_OPEN_APROPOS][5] = 0;
config.keys[PA_OPEN_APROPOS][6] = 0;
config.keys[PA_OPEN_APROPOS][7] = 0;

// Default value for: Perform whatis on the focused link
config.keys[PA_OPEN_WHATIS][0] = wcs2ch(L"w");
config.keys[PA_OPEN_WHATIS][1] = 0;
config.keys[PA_OPEN_WHATIS][2] = 0;
config.keys[PA_OPEN_WHATIS][3] = 0;
config.keys[PA_OPEN_WHATIS][4] = 0;
config.keys[PA_OPEN_WHATIS][5] = 0;
config.keys[PA_OPEN_WHATIS][6] = 0;
config.keys[PA_OPEN_WHATIS][7] = 0;

// Default value for: Open a manual page using a dialog
config.keys[PA_SP_OPEN][0] = wcs2ch(L"O");
config.keys[PA_SP_OPEN][1] = 0;
config.keys[PA_SP_OPEN][2] = 0;
config.keys[PA_SP_OPEN][3] = 0;
config.keys[PA_SP_OPEN][4] = 0;
config.keys[PA_SP_OPEN][5] = 0;
config.keys[PA_SP_OPEN][6] = 0;
config.keys[PA_SP_OPEN][7] = 0;

// Default value for: Perform apropos using a dialog
config.keys[PA_SP_APROPOS][0] = wcs2ch(L"A");
config.keys[PA_SP_APROPOS][1] = 0;
config.keys[PA_SP_APROPOS][2] = 0;
config.keys[PA_SP_APROPOS][3] = 0;
config.keys[PA_SP_APROPOS][4] = 0;
config.keys[PA_SP_APROPOS][5] = 0;
config.keys[PA_SP_APROPOS][6] = 0;
config.keys[PA_SP_APROPOS][7] = 0;

// Default value for: Perform whatis using a dialog
config.keys[PA_SP_WHATIS][0] = wcs2ch(L"W");
config.keys[PA_SP_WHATIS][1] = 0;
config.keys[PA_SP_WHATIS][2] = 0;
config.keys[PA_SP_WHATIS][3] = 0;
config.keys[PA_SP_WHATIS][4] = 0;
config.keys[PA_SP_WHATIS][5] = 0;
config.keys[PA_SP_WHATIS][6] = 0;
config.keys[PA_SP_WHATIS][7] = 0;

// Default value for: Go to index (home) page
config.keys[PA_INDEX][0] = wcs2ch(L"i");
config.keys[PA_INDEX][1] = wcs2ch(L"I");
config.keys[PA_INDEX][2] = 0;
config.keys[PA_INDEX][3] = 0;
config.keys[PA_INDEX][4] = 0;
config.keys[PA_INDEX][5] = 0;
config.keys[PA_INDEX][6] = 0;
config.keys[PA_INDEX][7] = 0;

// Default value for: Go back one step in history
config.keys[PA_BACK][0] = wcs2ch(L"KEY_BACKSPACE");
config.keys[PA_BACK][1] = wcs2ch(L"BS");
config.keys[PA_BACK][2] = wcs2ch(L"[");
config.keys[PA_BACK][3] = 0;
config.keys[PA_BACK][4] = 0;
config.keys[PA_BACK][5] = 0;
config.keys[PA_BACK][6] = 0;
config.keys[PA_BACK][7] = 0;

// Default value for: Go forward one step in history
config.keys[PA_FWRD][0] = wcs2ch(L"]");
config.keys[PA_FWRD][1] = 0;
config.keys[PA_FWRD][2] = 0;
config.keys[PA_FWRD][3] = 0;
config.keys[PA_FWRD][4] = 0;
config.keys[PA_FWRD][5] = 0;
config.keys[PA_FWRD][6] = 0;
config.keys[PA_FWRD][7] = 0;

// Default value for: Show history menu
config.keys[PA_HISTORY][0] = wcs2ch(L"s");
config.keys[PA_HISTORY][1] = wcs2ch(L"S");
config.keys[PA_HISTORY][2] = 0;
config.keys[PA_HISTORY][3] = 0;
config.keys[PA_HISTORY][4] = 0;
config.keys[PA_HISTORY][5] = 0;
config.keys[PA_HISTORY][6] = 0;
config.keys[PA_HISTORY][7] = 0;

// Default value for: Show table of contents
config.keys[PA_TOC][0] = wcs2ch(L"t");
config.keys[PA_TOC][1] = wcs2ch(L"T");
config.keys[PA_TOC][2] = 0;
config.keys[PA_TOC][3] = 0;
config.keys[PA_TOC][4] = 0;
config.keys[PA_TOC][5] = 0;
config.keys[PA_TOC][6] = 0;
config.keys[PA_TOC][7] = 0;

// Default value for: Forward search
config.keys[PA_SEARCH][0] = wcs2ch(L"/");
config.keys[PA_SEARCH][1] = 0;
config.keys[PA_SEARCH][2] = 0;
config.keys[PA_SEARCH][3] = 0;
config.keys[PA_SEARCH][4] = 0;
config.keys[PA_SEARCH][5] = 0;
config.keys[PA_SEARCH][6] = 0;
config.keys[PA_SEARCH][7] = 0;

// Default value for: Backward search
config.keys[PA_SEARCH_BACK][0] = wcs2ch(L"?");
config.keys[PA_SEARCH_BACK][1] = 0;
config.keys[PA_SEARCH_BACK][2] = 0;
config.keys[PA_SEARCH_BACK][3] = 0;
config.keys[PA_SEARCH_BACK][4] = 0;
config.keys[PA_SEARCH_BACK][5] = 0;
config.keys[PA_SEARCH_BACK][6] = 0;
config.keys[PA_SEARCH_BACK][7] = 0;

// Default value for: Go to next search result
config.keys[PA_SEARCH_NEXT][0] = wcs2ch(L"n");
config.keys[PA_SEARCH_NEXT][1] = 0;
config.keys[PA_SEARCH_NEXT][2] = 0;
config.keys[PA_SEARCH_NEXT][3] = 0;
config.keys[PA_SEARCH_NEXT][4] = 0;
config.keys[PA_SEARCH_NEXT][5] = 0;
config.keys[PA_SEARCH_NEXT][6] = 0;
config.keys[PA_SEARCH_NEXT][7] = 0;

// Default value for: Go to previous search results
config.keys[PA_SEARCH_PREV][0] = wcs2ch(L"N");
config.keys[PA_SEARCH_PREV][1] = 0;
config.keys[PA_SEARCH_PREV][2] = 0;
config.keys[PA_SEARCH_PREV][3] = 0;
config.keys[PA_SEARCH_PREV][4] = 0;
config.keys[PA_SEARCH_PREV][5] = 0;
config.keys[PA_SEARCH_PREV][6] = 0;
config.keys[PA_SEARCH_PREV][7] = 0;

// Default value for: Show this help message
config.keys[PA_HELP][0] = wcs2ch(L"h");
config.keys[PA_HELP][1] = wcs2ch(L"H");
config.keys[PA_HELP][2] = 0;
config.keys[PA_HELP][3] = 0;
config.keys[PA_HELP][4] = 0;
config.keys[PA_HELP][5] = 0;
config.keys[PA_HELP][6] = 0;
config.keys[PA_HELP][7] = 0;

// Default value for: Exit the program
config.keys[PA_QUIT][0] = wcs2ch(L"q");
config.keys[PA_QUIT][1] = wcs2ch(L"Q");
config.keys[PA_QUIT][2] = 0;
config.keys[PA_QUIT][3] = 0;
config.keys[PA_QUIT][4] = 0;
config.keys[PA_QUIT][5] = 0;
config.keys[PA_QUIT][6] = 0;
config.keys[PA_QUIT][7] = 0;

// Default value for: Enable mouse support
config.mouse.enable = wcs2bool(L"false");

// Default value for: Swap the left and right mouse buttons
config.mouse.left_handed = wcs2bool(L"false");

// Default value for: Causes the left mouse button to invoke the OPEN action and/or act as the ENTER key
config.mouse.left_click_open = wcs2bool(L"false");

// Default value for: True if we are in TUI mode, false if we are in CLI mode
config.layout.tui = wcs2bool(L"true");

// Default value for: If true, don't change width to match the terminal size
config.layout.fixedwidth = wcs2bool(L"true");

// Default value for: Current terminal width
config.layout.width = wcs2int(L"80");

// Default value for: Current terminal height
config.layout.height = wcs2int(L"25");

// Default value for: Scrollbar width
config.layout.sbar_width = wcs2int(L"1");

// Default value for: Status bar height
config.layout.stat_height = wcs2int(L"2");

// Default value for: Main window width
config.layout.main_width = wcs2int(L"79");

// Default value for: Main window height
config.layout.main_height = wcs2int(L"23");

// Default value for: Narrow pop-up dialogs width
config.layout.imm_width_narrow = wcs2int(L"54");

// Default value for: Wide pop-up dialogs width
config.layout.imm_width_wide = wcs2int(L"54");

// Default value for: Short pop-up dialogs height
config.layout.imm_height_short = wcs2int(L"6");

// Default value for: Long pop-up dialogs height
config.layout.imm_height_long = wcs2int(L"10");

// Default value for: If true, show a list of sections at the top of each page
config.layout.sections_on_top = wcs2bool(L"true");

// Default value for: Left margin size
config.layout.lmargin = wcs2int(L"2");

// Default value for: Right margin size
config.layout.rmargin = wcs2int(L"2");

// Default value for: Number of characters in a tab stop
config.layout.tabstop = wcs2int(L"8");

// Default value for: If true, show the scrollbar
config.layout.sbar = wcs2bool(L"true");

// Default value for: If true, beep the terminal
config.layout.beep = wcs2bool(L"true");

// Default value for: Program executable basename (discovered automatically)
config.misc.program_name = NULL;

// Default value for: Formal program name and version
config.misc.program_version = xwcsdup(L"Qman 1.4.1");

// Default value for: Path to the configuration file
config.misc.config_path = NULL;

// Default value for: Path to the man(1) command
config.misc.man_path = xstrdup("/usr/bin/man");

// Default value for: Path to the groff(1) command
config.misc.groff_path = xstrdup("/usr/bin/groff");

// Default value for: Path to the whatis(1) command
config.misc.whatis_path = xstrdup("/usr/bin/whatis");

// Default value for: Path to the apropos(1) command
config.misc.apropos_path = xstrdup("/usr/bin/apropos");

// Default value for: Path to web browser command
config.misc.browser_path = xstrdup("/usr/bin/xdg-open");

// Default value for: Path to mailer command
config.misc.mailer_path = xstrdup("/usr/bin/xdg-email");

// Default value for: Re-initialize curses after opening an http(s) link
config.misc.reset_after_http = wcs2bool(L"true");

// Default value for: Re-initialize curses after opening an e-mail link
config.misc.reset_after_email = wcs2bool(L"true");

// Default value for: Maximum number of history entries
config.misc.history_size = wcs2int(L"65536");

// Default value for: Hyphenate long words in manual pages
config.misc.hyphenate = wcs2bool(L"true");

// Default value for: Justify manual pages text
config.misc.justify = wcs2bool(L"true");

// Default value for: -a / --all option was passed
config.misc.global_whatis = wcs2bool(L"false");

// Default value for: -k / --global-whatis option was passed
config.misc.global_apropos = wcs2bool(L"false");


  if (NULL == config.misc.program_name)
    config.misc.program_name = program_invocation_short_name;

  set_transition_colours;
}

void configure() {
  static bool ft = true;    // true if this is the first invocation of
                            // `configure()`
  char path[BS_LINE];       // path to config file
  FILE *fp = NULL;          // file pointer to config file
  wchar_t errmsg[BS_SHORT]; // error message

  wcs2colour(NULL);

  // If this isn't the first invocation of `configure()`, call
  // `conf_init_chars_colours()` to reset UI characters and colors to their
  // default values
  if (ft)
    ft = false;
  else
    conf_init_chars_colours();

  // Find and open the config file
  if (NULL != config.misc.config_path) {
    // `-C` option was used; try to open specified config file, and fail if not
    // possible
    xstrncpy(path, config.misc.config_path, BS_LINE);
    fp = fopen(path, "r");
    if (NULL == fp) {
      swprintf(errmsg, BS_SHORT, L"Unable to open config file '%s'",
               config.misc.config_path);
      winddown(ES_CONFIG_ERROR, errmsg);
    }
  } else {
    // No `-C` option; try to find and open a config file at the standard
    // locations

    // ${XCD_CONFIG_HOME}/qman/qman.conf
    const char *xch = getenv("XDG_CONFIG_HOME");
    if (NULL != xch) {
      snprintf(path, BS_LINE, "%s/qman/qman.conf", xch);
      fp = fopen(path, "r");
    }

    // ${HOME}/.config/qman/qman.conf
    if (NULL == fp) {
      const char *home = getenv("HOME");
      if (NULL != home) {
        snprintf(path, BS_LINE, "%s/.config/qman/qman.conf", home);
        fp = fopen(path, "r");
      }
    }

    // ${P}/qman/qman.conf, where ${P} is a path in ${XDG_CONFIG_DIRS}
    if (NULL == fp) {
      const char *xcd = getenv("XDG_CONFIG_DIRS");
      if (NULL != xcd) {
        char *xcd_dup = xstrdup(xcd);
        char **xcd_paths = xcalloc(BS_SHORT, sizeof(char *));
        unsigned xcd_count = split_path(&xcd_paths, BS_SHORT, xcd_dup);
        for (unsigned i = 0; i < xcd_count; i++) {
          snprintf(path, BS_LINE, "%s/qman/qman.conf", xcd_paths[i]);
          fp = fopen(path, "r");
          if (NULL != fp)
            break;
        }
        free(xcd_dup);
        free(xcd_paths);
      }
    }

    // /etc/xdg/qman/qman.conf
    if (NULL == fp) {
      xstrncpy(path, "/etc/xdg/qman/qman.conf", BS_LINE);
      fp = fopen(path, "r");
    }

    // /etc/qman/qman.conf
    if (NULL == fp) {
      xstrncpy(path, "/etc/qman/qman.conf", BS_LINE);
      fp = fopen(path, "r");
    }
  }

  // If we have a config file, process it
  if (NULL != fp) {
    fclose(fp);
    eini_init();
    eini(&conf_handler, &error_handler, path);
    eini_winddown();
  }
  set_transition_colours;
}
