// Configuration and configuration file handling (definition)

#ifndef CONFIG_H

#define CONFIG_H

#include "lib.h"

//
// Types
//

// A ternary bit, i.e. one that can take one of 3 values: `true`, `false`, or
// `auto`
typedef enum {
  t_false,
  t_true,
  t_auto
} trit_t;

// The color of a text character. Unlike ncurses colors, which are short
// integers that correspond to a color value, `colour_t` combines foreground
// color, background color, boldness, and a corresponding initialized ncurses
// color pair.
typedef struct {
  short fg;   // foreground ncurses color no.
  bool bold;  // true if foreground color is bold, false otherwise
  short bg;   // background ncurses color no.
  short pair; // ncurses color pair no.
} colour_t;

// A program action
typedef enum {
  PA_NULL,                       // Do nothing (must be first member of `action_t`)
  PA_UP,                         // Scroll up one line
  PA_DOWN,                       // Scroll down one line
  PA_LEFT,                       // Scroll left one tab stop
  PA_RIGHT,                      // Scroll right one tab stop
  PA_PGUP,                       // Scroll up one page
  PA_PGDN,                       // Scroll down one page
  PA_HOME,                       // Go to page top
  PA_END,                        // Go to page end
  PA_OPEN,                       // Open focused link
  PA_OPEN_APROPOS,               // Perform apropos on the focused link
  PA_OPEN_WHATIS,                // Perform whatis on the focused link
  PA_SP_OPEN,                    // Open a manual page using a dialog
  PA_SP_APROPOS,                 // Perform apropos using a dialog
  PA_SP_WHATIS,                  // Perform whatis using a dialog
  PA_INDEX,                      // Go to index (home) page
  PA_BACK,                       // Go back one step in history
  PA_FWRD,                       // Go forward one step in history
  PA_HISTORY,                    // Show history menu
  PA_TOC,                        // Show table of contents
  PA_SEARCH,                     // Forward search
  PA_SEARCH_BACK,                // Backward search
  PA_SEARCH_NEXT,                // Go to next search result
  PA_SEARCH_PREV,                // Go to previous search results
  PA_HELP,                       // Show this help message
  PA_QUIT                        // Exit the program
} action_t;

// Configuration structure for section `tcap`
typedef struct {
  // Number of colors supported by the terminal (set to -1 to auto-configure)
  unsigned colours;

  // Terminal colors can be re-defined
  trit_t rgb;

  // Terminal supports Unicode
  trit_t unicode;

  // Terminal supports clipboard operations (OSC 52)
  trit_t clipboard;

  // Terminal escape delay
  unsigned escdelay;
} config_tcap_t;

// Configuration structure for section `chars`
typedef struct {
  // Scrollbar top end
  wchar_t *sbar_top;

  // Scrollbar track line
  wchar_t *sbar_vline;

  // Scrollbar bottom end
  wchar_t *sbar_bottom;

  // Scrollbar knob
  wchar_t *sbar_block;

  // Transition between the mode to name sections of the status bar
  wchar_t *trans_mode_name;

  // Transition between the name and location sections of the status bar
  wchar_t *trans_name_loc;

  // Transition between the prompt and help sections of the status bar
  wchar_t *trans_prompt_help;

  // Transition between the prompt and error message sections of the status bar
  wchar_t *trans_prompt_em;

  // Dialog box horizontal line
  wchar_t *box_hline;

  // Dialog box vertical line
  wchar_t *box_vline;

  // Dialog box top left corner
  wchar_t *box_tl;

  // Dialog box top right corner
  wchar_t *box_tr;

  // Dialog box bottom left corner
  wchar_t *box_bl;

  // Dialog box bottom right corner
  wchar_t *box_br;

  // Up arrow
  wchar_t *arrow_up;

  // Down arrow
  wchar_t *arrow_down;
} config_chars_t;

// Configuration structure for section `colours`
typedef struct {
  // Fallback for B&W terminals
  colour_t fallback;

  // Page text
  colour_t text;

  // Matched search terms in page text
  colour_t search;

  // Marked text
  colour_t mark;

  // Links to manual pages
  colour_t link_man;

  // Links to manual pages (focused)
  colour_t link_man_f;

  // HTTP links
  colour_t link_http;

  // HTTP links (focused)
  colour_t link_http_f;

  // E-mail links
  colour_t link_email;

  // E-mail links (focused)
  colour_t link_email_f;

  // In-page local search links
  colour_t link_ls;

  // In-page local search links (focused)
  colour_t link_ls_f;

  // Scrollbar track line
  colour_t sb_line;

  // Scrollbar knob
  colour_t sb_block;

  // Status bar mode section
  colour_t stat_indic_mode;

  // Status bar name section
  colour_t stat_indic_name;

  // Status bar location section
  colour_t stat_indic_loc;

  // Status bar input prompt
  colour_t stat_input_prompt;

  // Status bar help section
  colour_t stat_input_help;

  // Status bar error message section
  colour_t stat_input_em;

  // Pop-up dialogs border
  colour_t imm_border;

  // Pop-up dialogs title
  colour_t imm_title;

  // Pop-up input dialog prompt
  colour_t sp_input;

  // Pop-up input dialog progressive help text
  colour_t sp_text;

  // Pop-up input dialog progressive help text (focused)
  colour_t sp_text_f;

  // Help dialog entries text
  colour_t help_text;

  // Help dialog entries text (focused)
  colour_t help_text_f;

  // History dialog entries text
  colour_t history_text;

  // History dialog entries text (focused)
  colour_t history_text_f;

  // TOC dialog entries text
  colour_t toc_text;

  // TOC dialog entries text (focused)
  colour_t toc_text_f;

  // Color pair for mode to name transition character
  unsigned trans_mode_name;

  // Color pair for name to location transition character
  unsigned trans_name_loc;

  // Color pair for prompt to help transition character
  unsigned trans_prompt_help;

  // Color pair for prompt to error_message transition character
  unsigned trans_prompt_em;
} config_colours_t;

// Configuration structure for section `keys`
typedef int config_keys_t[PA_QUIT + 1][8];

// Configuration structure for section `mouse`
typedef struct {
  // Enable mouse support
  bool enable;

  // Swap the left and right mouse buttons
  bool left_handed;

  // Causes the left mouse button to invoke the OPEN action and/or act as the ENTER key
  bool left_click_open;
} config_mouse_t;

// Configuration structure for section `layout`
typedef struct {
  // True if we are in TUI mode, false if we are in CLI mode
  bool tui;

  // If true, don't change width to match the terminal size
  bool fixedwidth;

  // Current terminal width
  unsigned width;

  // Current terminal height
  unsigned height;

  // Scrollbar width
  unsigned sbar_width;

  // Status bar height
  unsigned stat_height;

  // Main window width
  unsigned main_width;

  // Main window height
  unsigned main_height;

  // Narrow pop-up dialogs width
  unsigned imm_width_narrow;

  // Wide pop-up dialogs width
  unsigned imm_width_wide;

  // Short pop-up dialogs height
  unsigned imm_height_short;

  // Long pop-up dialogs height
  unsigned imm_height_long;

  // If true, show a list of sections at the top of each page
  bool sections_on_top;

  // Left margin size
  unsigned lmargin;

  // Right margin size
  unsigned rmargin;

  // Number of characters in a tab stop
  unsigned tabstop;

  // If true, show the scrollbar
  bool sbar;

  // If true, beep the terminal
  bool beep;
} config_layout_t;

// Configuration structure for section `misc`
typedef struct {
  // Program executable basename (discovered automatically)
  char *program_name;

  // Formal program name and version
  wchar_t *program_version;

  // Path to the configuration file
  char *config_path;

  // Path to the man(1) command
  char *man_path;

  // Path to the groff(1) command
  char *groff_path;

  // Path to the whatis(1) command
  char *whatis_path;

  // Path to the apropos(1) command
  char *apropos_path;

  // Path to web browser command
  char *browser_path;

  // Path to mailer command
  char *mailer_path;

  // Re-initialize curses after opening an http(s) link
  bool reset_after_http;

  // Re-initialize curses after opening an e-mail link
  bool reset_after_email;

  // Maximum number of history entries
  unsigned history_size;

  // Hyphenate long words in manual pages
  bool hyphenate;

  // Justify manual pages text
  bool justify;

  // -a / --all option was passed
  bool global_whatis;

  // -k / --global-whatis option was passed
  bool global_apropos;
} config_misc_t;

// Main configuration structure
typedef struct {
  config_tcap_t tcap;
  config_chars_t chars;
  config_colours_t colours;
  config_keys_t keys;
  config_mouse_t mouse;
  config_layout_t layout;
  config_misc_t misc;
} config_t;

//
// Constants
//

// Alternate configuration file locations
#define CONFIG_PATHS [ "~/.config/qman.conf", "/etc/xdg/qman.conf" ]

//
// Global variables
//

// Program configuration
extern config_t config;

// String representations for program actions
extern const char *keys_names[PA_QUIT + 1];

// Help text for program actions and their corresponding key character mappings
extern const wchar_t *keys_help[PA_QUIT + 1];

//
// Functions
//

// Initialize configuration to sane defaults
extern void conf_init();

// Locate and parse the configuration file, setting all members of `config`
// appropriately
extern void configure();

#endif
